

# fitting the model
# data is a list that must contain: n, y, which_in_likelihood, n_in_likelihood, model_number, L
# the stan-model
# use_likelihood: 0 for sampling from the prior, 1 for sampling from the posterior
# n_burnin: number of samples to burn-in
# n_samp: number of samples AFTER burn-in (total number of samples = n_burnin + n_samp)
fit_stan <- function(data, stan_mod, use_likelihood = 1, n_burnin = 2500, n_samp = 5000){
  
  stan_data <- data
  save_pars <- c("theta", "intercept", "eta", "genetic_effects", "log_pred_dens")
  stan_data$use_likelihood <- use_likelihood
  
  res_stan <- sampling(
    object = stan_mod,
    data = stan_data,
    seed = 1, # making the simulations reproducable
    warmup = n_burnin,
    iter = n_burnin + n_samp,
    thin = 5, # do not need to store all samples
    pars = save_pars,
    init = "0", # initialize all parameters to zero
    control = list(adapt_delta = 0.95),
    chains = 1
  )
  
  return(res_stan)
  
}


# fit the model with maximum likelihood
# data is a list that must contain: y, which_in_likelihood, model_number, startvalues, V_A, V_D, V_X
fit_ML <- function(data){
  
  model_number <- data$model_number # 1 = A-MW-exp, 2 = AD-MW-exp, 3 = ADX-MW-exp
  
  if (model_number == 1){
    opt_func <- function(x) likelihood(pars = c(x[1], x[2], -Inf, -Inf, x[3]), data)
    grad_func <- function(x) gradient(pars = c(x[1], x[2], -Inf, -Inf, x[3]), data)[c(1,2,5)]
    n_param <- 3
  } else if (model_number == 2){
    opt_func <- function(x) likelihood(pars = c(x[1], x[2], x[3], -Inf, x[4]), data)
    grad_func <- function(x) gradient(pars = c(x[1], x[2], x[3], -Inf, x[4]), data)[c(1,2,3,5)]
    n_param <- 4
  } else { # model_number == 3
    opt_func <- function(x) likelihood(pars = x, data)
    grad_func <- function(x) gradient(pars = x, data)
    n_param <- 5
  }
  
  # exp(-30) > 1e-16
  res <- nloptr(
    # x0 = rep(0, n_param),
    x0 = data$startvalues,
    eval_f = opt_func,
    eval_grad_f = grad_func,
    lb = c(-100, rep(-50, n_param-2), -20), # to avoid singlar covariance matrix, the residual variance cannot be too low (but it never should be anyway)
    ub = c(100, rep(20, n_param-1)),
    opts = list(algorithm = "NLOPT_LD_LBFGS",
                check_derivatives_tol = 1e-5,
                maxeval = 1e4,
                print_level = 1)
  )
  
  return(res)
  
}

# pars: mu, logvars (mu, a_var, d_var, x_var, e_var)
likelihood <- function(pars, data){
  
  y <- data$y
  n <- length(y)
  # have scaled matrices to have typical variance 1 from before
  V_a <- data$V_A
  V_d <- data$V_D
  V_e <- data$V_X
  
  # remove the missing data and data that is excluded in the CV
  inc <- data$which_in_likelihood
  
  y <- y[inc]
  n <- length(y)
  V_a <- V_a[inc, inc]
  V_d <- V_d[inc, inc]
  V_e <- V_e[inc, inc]
  
  eta <- matrix(pars[1]*rep(1, n), ncol = 1)
  
  cov_mat <- 
    exp(pars[2]) * V_a +
    exp(pars[3]) * V_d +
    exp(pars[4]) * V_e +
    exp(pars[5]) * diag(n)
  
  loglik <- -0.5*determinant(cov_mat, logarithm = TRUE)$modulus - 0.5 * t(y-eta) %*% solve(cov_mat) %*% (y-eta)
  
  return(-as.numeric(loglik))
  
}

# pars: mu, logvars
gradient <- function(pars, data){
  
  y <- data$y
  n <- length(y)
  V_a <- data$V_A
  V_d <- data$V_D
  V_e <- data$V_X
  
  # remove the missing data and data that is excluded in the CV
  inc <- data$which_in_likelihood
  
  y <- y[inc]
  n <- length(y)
  V_a <- V_a[inc, inc]
  V_d <- V_d[inc, inc]
  V_e <- V_e[inc, inc]
  
  eta <- pars[1]*rep(1, n)
  
  cov_mat <- 
    exp(pars[2]) * V_a +
    exp(pars[3]) * V_d +
    exp(pars[4]) * V_e +
    exp(pars[5]) * diag(n) 
  
  ## calculating stuff we need many times
  d_sig_a <- exp(pars[2]) * V_a
  d_sig_d <- exp(pars[3]) * V_d
  d_sig_e <- exp(pars[4]) * V_e
  d_sig_res <- exp(pars[5]) * diag(n)
  prec_mat <- solve(cov_mat)
  tmp <- prec_mat %*% (y - eta)
  
  deriv1 <- sum(tmp)
  deriv2 <- -0.5 * sum(diag(prec_mat %*% d_sig_a)) + 0.5 * (t(tmp) %*% d_sig_a %*% tmp)
  deriv3 <- -0.5 * sum(diag(prec_mat %*% d_sig_d)) + 0.5 * (t(tmp) %*% d_sig_d %*% tmp)
  deriv4 <- -0.5 * sum(diag(prec_mat %*% d_sig_e)) + 0.5 * (t(tmp) %*% d_sig_e %*% tmp)
  deriv5 <- -0.5 * sum(diag(prec_mat %*% d_sig_res)) + 0.5 * (t(tmp) %*% d_sig_res %*% tmp)
  
  return(-c(deriv1, deriv2, deriv3, deriv4, deriv5))
  
}


