
## All the necessary functions for running the example is included

fit_model <- function(model_number, ...){
  
  model_name <- c(
    "A-comp",
    "A-comp*",
    "A-tree",
    "A-tree*",
    "A-ML",
    "AD-comp",
    "AD-comp*",
    "AD-tree",
    "AD-tree*",
    "AD-ML",
    "ADX-comp",
    "ADX-comp*",
    "ADX-tree",
    "ADX-tree*",
    "ADX-tree-opp",
    "ADX-ML",
    "Phenotype selection"
  )
  
  modname <- model_name[model_number]
  
  cat("\n\nFitting ", modname, ", which is model number ", model_number, "\n\n", sep = "")
  
  mydata <- readRDS(paste0("data/data_model_", model_number, ".rds"))
  
  if (model_number == 17){ # Phenotype selection
    
    res2 <- mydata$stan_data
    res2$model_type <- "phenotype"
    res2$model <- mydata$data$model
    
  } else if (model_number %in% c(5, 10, 16)){ # Maximum likelihood
    
    res <- fit_ML(mydata)
    
    res2 <- res
    res2$model_type <- "ml"
    res2$model <- mydata$data$model
    
  } else { # Stan
    
    res <- fit_stan(mydata, ...)
    
    res2 <- rstan::extract(res)
    res2$model_type <- "stan"
    res2$model <- mydata$data$model
    res2$model_number <- model_number
    
  }
  
  return(res2)
  
  
}

# for fitting models in stan, input is the data, how many samples (choose between low, medium and high),
# and whether to use the likelihood (1) or not (0, simulate from the prior)
fit_stan <- function(data, no_of_samps = c("medium", "low", "high"), use_likelihood = 1){
  
  no_of_samps <- match.arg(no_of_samps)
  
  mod_name <- if (length(data$data$params_to_save) == 5) "stan_models/wheat_simple.stan" else 
    if (length(data$data$params_to_save) == 6) "stan_models/wheat_ad.stan" else
      "stan_models/wheat.stan"
  
  stan_mod <- stan_model(file = mod_name, auto_write = TRUE)
  
  stan_data <- data$stan_data
  stan_data$use_likelihood <- use_likelihood
  
  if (no_of_samps == "low"){
    no_warmup <- 100
    no_iter <- 200
  } else if (no_of_samps == "medium"){
    no_warmup <- 2000
    no_iter <- 4000
  } else {
    no_warmup <- 20000
    no_iter <- 40000
  }
  
  res_stan <- sampling(
    object = stan_mod,
    data = stan_data,
    seed = 1, # making the simulations reproducable
    warmup = no_warmup,
    iter = no_iter,
    pars = data$data$params_to_save,
    init = "0", # initialize all parameters to zero
    control = list(adapt_delta = 0.95),
    chains = 1
  )
  
  return(res_stan)
  
}

# fot fitting models with maximum likelihood, input is the data
fit_ML <- function(data){
  
  mod <- data$data$model
  
  if (mod == "small"){ # no dominance and epistasis variance
    opt_func <- function(x) likelihood(pars = c(x[1], x[2], -Inf, -Inf, x[3]), data$stan_data)
    grad_func <- function(x) gradient(pars = c(x[1], x[2], -Inf, -Inf, x[3]), data)[c(1,2,5)]
    n_param <- 3
  } else if (mod == "medium"){
    opt_func <- function(x) likelihood(pars = c(x[1], x[2], x[3], -Inf, x[4]), data$stan_data)
    grad_func <- function(x) gradient(pars = c(x[1], x[2], x[3], -Inf, x[4]), data$stan_data)[c(1,2,3,5)]
    n_param <- 4
  } else {
    opt_func <- function(x) likelihood(pars = x, data$stan_data)
    grad_func <- function(x) gradient(pars = x, data$stan_data)
    n_param <- 5
  }
  
  # exp(-30) > 1e-16
  res_ML <- nloptr(
    rep(0, n_param),
    eval_f = opt_func,
    eval_grad_f = grad_func,
    lb = c(-100, rep(-50, n_param-2), -20), # to avoid singlar covariance matrix, the residual variance cannot be too low (but it never should be anyway)
    ub = c(100, rep(20, n_param-1)),
    opts = list(algorithm = "NLOPT_LD_LBFGS",
                check_derivatives_tol = 1e-5,
                maxeval = 1e4)
  )
  
  ## calculate the random effects
  sigma_ay <- exp(res_ML$solution[2]) * data$stan_data$V_a
  sigma_dy <- if (mod == "big" || mod == "medium") exp(res_ML$solution[3]) * data$stan_data$V_d else matrix(0, nrow = nrow(data$stan_data$V_d), ncol = ncol(data$stan_data$V_d))
  sigma_ey <- if (mod == "big") exp(res_ML$solution[4]) * data$stan_data$V_e else matrix(0, nrow = nrow(data$stan_data$V_e), ncol = ncol(data$stan_data$V_e))
  sigma_yy <- sigma_ay + sigma_dy + sigma_ey + exp(res_ML$solution[length(res_ML$solution)])*diag(nrow(data$stan_data$V_a))
  
  a_eff <- c(sigma_ay %*% solve(sigma_yy) %*% (data$stan_data$y - res_ML$solution[1]))
  d_eff <- c(sigma_dy %*% solve(sigma_yy) %*% (data$stan_data$y - res_ML$solution[1]))
  e_eff <- c(sigma_ey %*% solve(sigma_yy) %*% (data$stan_data$y - res_ML$solution[1]))
  
  var_a <- sigma_ay - sigma_ay %*% solve(sigma_yy) %*% t(sigma_ay)
  var_d <- sigma_dy - sigma_dy %*% solve(sigma_yy) %*% t(sigma_dy)
  var_e <- sigma_ey - sigma_ey %*% solve(sigma_yy) %*% t(sigma_ey)
  
  eta <- a_eff + d_eff + e_eff + res_ML$solution[1]
  eta_vars <- diag(var_a + var_d + var_e) # sum of the variances
  
  res_ml_all <- list()
  res_ml_all$a_effect <- matrix(a_eff, nrow = 1)
  res_ml_all$eta <- matrix(eta, nrow = 1)
  
  res_ml_all$optim_object <- res_ML
  
  return(res_ml_all)
  
}

# likelihood for maximum likelihood
# pars: mu, logvars
likelihood <- function(pars, data){
  
  y <- data$y
  n <- length(y)
  # scaled matrices to have typical variance 1
  V_a <- data$V_a/data$typvar_a
  V_d <- data$V_d/data$typvar_d
  V_e <- data$V_e/data$typvar_e
  
  eta <- matrix(pars[1]*rep(1, n), ncol = 1)
  
  cov_mat <- 
    exp(pars[2]) * V_a +
    exp(pars[3]) * V_d +
    exp(pars[4]) * V_e +
    exp(pars[5]) * diag(n)
  
  loglik <- -0.5*determinant(cov_mat, logarithm = TRUE)$modulus - 0.5 * t(y-eta) %*% solve(cov_mat) %*% (y-eta) # proportional to
  
  return(-as.numeric(loglik))
  
}

# gradient for maximum likelihood
# pars: mu, logvars
gradient <- function(pars, data){
  
  y <- data$y
  n <- length(y)
  V_a <- data$V_a/data$typvar_a
  V_d <- data$V_d/data$typvar_d
  V_e <- data$V_e/data$typvar_e
  
  eta <- pars[1]*rep(1, n)
  
  cov_mat <- 
    exp(pars[2]) * V_a +
    exp(pars[3]) * V_d +
    exp(pars[4]) * V_e +
    exp(pars[5]) * diag(n) 
  
  ## calculating stuff we need many times
  d_sig_a <- exp(pars[2]) * V_a
  d_sig_d <- exp(pars[3]) * V_d
  d_sig_e <- exp(pars[4]) * V_e
  d_sig_res <- exp(pars[5]) * diag(n)
  prec_mat <- solve(cov_mat)
  tmp <- prec_mat %*% (y - eta)
  
  deriv1 <- sum(tmp)
  deriv2 <- -0.5 * sum(diag(prec_mat %*% d_sig_a)) + 0.5 * (t(tmp) %*% d_sig_a %*% tmp)
  deriv3 <- -0.5 * sum(diag(prec_mat %*% d_sig_d)) + 0.5 * (t(tmp) %*% d_sig_d %*% tmp)
  deriv4 <- -0.5 * sum(diag(prec_mat %*% d_sig_e)) + 0.5 * (t(tmp) %*% d_sig_e %*% tmp)
  deriv5 <- -0.5 * sum(diag(prec_mat %*% d_sig_res)) + 0.5 * (t(tmp) %*% d_sig_res %*% tmp)
  
  return(-c(deriv1, deriv2, deriv3, deriv4, deriv5))
  
}




# plotting the histograms of the posterior variance proportions
plot_variance_proportions <- function(res_obj){
  
  if (res_obj$model_type != "stan") return("Can only plot the histogram of a model fittet using Stan.")
  
  if (ncol(res_obj$theta_unscaled) == 2){
    
    totvar <- rowSums(exp(res_obj$theta_unscaled))
    w_g <- exp(res_obj$theta_unscaled[,2])/totvar
    w_1 <- NA
    w_2 <- NA
    
    df <- data.frame(
      w_g = w_g
    )
    
    names(df) <- c("Proportion of total variance \nto the genetic effect")
    
  } else if (ncol(res_obj$theta_unscaled) == 3) {
    
    totvar <- rowSums(exp(res_obj$theta_unscaled))
    w_g <- rowSums(exp(res_obj$theta_unscaled[,2:3]))/totvar
    w_1 <- exp(res_obj$theta_unscaled[,2])/rowSums(exp(res_obj$theta_unscaled[,2:3]))
    w_2 <- NA
    
    df <- data.frame(
      w_g = w_g, w_1 = w_1
    )
    
    names(df) <- c("Proportion of total variance \nto the genetic effect",
                   "Proportion of genetic variance \nto the additive effect")
    
  } else {
    
    totvar <- rowSums(exp(res_obj$theta_unscaled))
    w_g <- rowSums(exp(res_obj$theta_unscaled[,2:4]))/totvar
    w_1 <- rowSums(exp(res_obj$theta_unscaled[,2]))/rowSums(exp(res_obj$theta_unscaled[,2:4]))
    w_2 <- exp(res_obj$theta_unscaled[,3])/rowSums(exp(res_obj$theta_unscaled[,3:4]))
    
    df <- data.frame(
      w_g = w_g, w_1 = w_1, w_2 = w_2
    )
    
    names(df) <- c("Proportion of total variance \nto the genetic effect",
                   "Proportion of genetic variance \nto the additive effect",
                   "Proportion of non-additive \nvariance to the dominance effect")
    
  }
  
  
  # prior is samples
  if (res_obj$model %in% c("A-comp", "A-comp*", "AD-comp", "AD-comp*", "ADX-comp", "ADX-comp*")){
    
    df$type <- "Posterior"
    
    df2 <- melt(df, id.vars = c("type"))
    
    gg <- ggplot(df2, aes(x = value, y = ..density.., fill = type, group = type)) +
      geom_histogram(position = "identity", alpha = 0.4, col = "black", bins = 40, na.rm = TRUE) +
      theme(legend.position = "bottom", legend.direction = "horizontal") +
      scale_fill_viridis_d(option = "plasma") +
      theme(legend.title = element_blank()) +
      xlab("Variance proportion") + ylab("Density") +
      ggtitle(paste0(res_obj$model, ", ", length(w_g), " samples"))
    
  } else {
    
    df2 <- melt(df, id.vars = c())
    
    gg <- ggplot() +
      geom_histogram(data = df2, mapping = aes(x = value, y = ..density.., fill = "Posterior"),
                     position = "identity", alpha = 0.4, col = "black", bins = 40, na.rm = TRUE) +
      theme(legend.position = "none") +
      scale_fill_viridis_d(option = "plasma") +
      scale_color_viridis_d(option = "plasma") +
      theme(legend.title = element_blank()) +
      xlab("Variance proportion") + ylab("Density") +
      ggtitle(paste0(res_obj$model, ", ", length(w_g), " samples"))
    
  }
  
  if (res_obj$model_number > 5){
    gg <- gg + facet_grid(~ variable)
  }
  
  return(gg)
  
}



find_ranking <- function(res_obj){
  
  capture.output(truth <- fit_model(17), file = "NULL")
  
  true_gen <- order(truth$gen_eff, decreasing = TRUE)
  true_add <- order(truth$a_eff, decreasing = TRUE)
  
  # genetic ranking
  if (res_obj$model_type == "phenotype"){
    est_gen <- order(res_obj$y, decreasing = TRUE)
    est_add <- order(res_obj$y, decreasing = TRUE)
  } else {
    est_gen <- order(colMeans(res_obj$eta), decreasing = TRUE)
    est_add <- order(colMeans(res_obj$a_effect), decreasing = TRUE)
  }
  
  gen_10 <- which(true_gen %in% est_gen[1:10])
  add_10 <- which(true_add %in% est_add[1:10])
  
  cat("\n", res_obj$model, "\n", sep = "")
  cat("---------------------\n")
  cat("Ranking of genetic effects: ", paste0(gen_10, sep = ", "), "which is ", sum(gen_10 <= 10), " out of 10 \n", sep = "")
  cat("Ranking of additive effects: ", paste0(add_10, sep = ", "), "which is ", sum(add_10 <= 10), " out of 10 \n", sep = "")
  cat("\n")
  
}










