

// Stan code for fitting A-MW-exp, AD-MW-exp, ADX-MW-exp

functions {
  
    // evaluates spline for one logit weight
    real eval_spline_lpdf(real logitw, vector knots, matrix[] prior_coeffs, int idx, int n_knots){
        real dx;
        real val;
        int idd;
        int low;
        int high;
        int med;
        
        // Search for correct cubic polynomial
        low = 1;
        high = n_knots;
        while(high-low > 1){
            med = (high+low)/2;
            if(logitw < knots[med]){
                high = med;
            } else{
                low = med;
            }
        }
        idd = low;
        dx = logitw-knots[idd];
        val =  prior_coeffs[idx, idd, 1];
        val += prior_coeffs[idx, idd, 2]*dx;
        val += prior_coeffs[idx, idd, 3]*dx*dx;
        val += prior_coeffs[idx, idd, 4]*dx*dx*dx;
    
        return(val);
    }
    
    // jeffreys prior
    real jeffrey_logvar_lpdf(real logvar){ // on log variance
      return 0;
    }
    
    // calculates the jacobian from V/w to log variances, and then takes the determinant, and log of the determinant for the three models
    
    // for AD
    real calc_log_jacdet_regtovar_ad(real[] theta){
      
      matrix[3,3] jacobian;
      real V;
      
      V = exp(theta[1]) + exp(theta[2]) + exp(theta[3]);
      
      jacobian[1,1] = exp(theta[1]);
      jacobian[1,2] = exp(theta[2]);
      jacobian[1,3] = exp(theta[3]);
      
      jacobian[2,1] = -exp(theta[1]) * (exp(theta[2]) + exp(theta[3])) / (V^2);
      jacobian[2,2] = (exp(theta[1]+theta[2])) / (V^2);
      jacobian[2,3] = (exp(theta[1]+theta[3])) / (V^2);
      
      jacobian[3,1] = 0;
      jacobian[3,2] = -(exp(theta[2] + theta[3])) / ((exp(theta[2]) + exp(theta[3]))^2);
      jacobian[3,3] = (exp(theta[2] + theta[3])) / ((exp(theta[2]) + exp(theta[3]))^2);
      
      return log(determinant(jacobian));
      
    }
    
    // for ADX
    real calc_log_jacdet_regtovar_adx(real[] theta){
      
      matrix[4,4] jacobian;
      real V;
      
      V = exp(theta[1]) + exp(theta[2]) + exp(theta[3]) + exp(theta[4]);
      
      jacobian[1,1] = exp(theta[1]);
      jacobian[1,2] = exp(theta[2]);
      jacobian[1,3] = exp(theta[3]);
      jacobian[1,4] = exp(theta[4]);
      
      jacobian[2,1] = -exp(theta[1]) * (exp(theta[2]) + exp(theta[3]) + exp(theta[4])) / (V^2);
      jacobian[2,2] = (exp(theta[1]+theta[2])) / (V^2);
      jacobian[2,3] = (exp(theta[1]+theta[3])) / (V^2);
      jacobian[2,4] = (exp(theta[1]+theta[4])) / (V^2);
      
      jacobian[3,1] = 0;
      jacobian[3,2] = -exp(theta[2]) * (exp(theta[3]) + exp(theta[4])) / ((exp(theta[2]) + exp(theta[3]) + exp(theta[4]))^2);
      jacobian[3,3] = (exp(theta[2]+theta[3])) / ((exp(theta[2]) + exp(theta[3]) + exp(theta[4]))^2);
      jacobian[3,4] = (exp(theta[2]+theta[4])) / ((exp(theta[2]) + exp(theta[3]) + exp(theta[4]))^2);
      
      jacobian[4,1] = 0;
      jacobian[4,2] = 0;
      jacobian[4,3] = -exp(theta[3] + theta[4]) / ((exp(theta[3]) + exp(theta[4]))^2);
      jacobian[4,4] = exp(theta[3] + theta[4]) / ((exp(theta[3]) + exp(theta[4]))^2);
      
      return log(determinant(jacobian));
      
    }
    
    // jacobian for A is in the prior function
    
    // calculates the joint prior for theta (not including covariates) for model A
    real prior_a_exp_lpdf(real[] theta, vector knots, matrix[] prior_coeffs, int n_knots, int use_likelihood){
      
      real logdens;
      real logV;
      real logitw;
      
      // initializing
      logdens = 0;
      
      logV = log(sum(exp(theta)));
      // note that logit(x/(x+y)) = log(x) - log(y)
      logitw = theta[2] - theta[1];
      
      // if we do not use the likelihood we must restrict the total variance since we use the jeffreys prior
      if (use_likelihood == 0){
        logdens += normal_lpdf(logV | 0, 1);
      } else {
        logdens += jeffrey_logvar_lpdf(logV); // V, total variance
      }
      
      // calculates the log determinant of the Jacobian of the transformation (log(V), logit(w)) -> (theta_1, theta_2)
      logdens += 0;
      
      // adding prior for the logit weights
      logdens += eval_spline_lpdf(logitw | knots, prior_coeffs, 1, n_knots); // w1, residual vs genetic/additive

      return(logdens);
      
    }
    
    // calculates the joint prior for theta (not including covariates) for model AD
    real prior_ad_exp_lpdf(real[] theta, vector knots, matrix[] prior_coeffs, int n_knots, int use_likelihood){
      
      real logdens;
      real logV;
      real logitw1; // amount of total to genetic variance
      real logitw2; // amount of genetic to dominance variance
      
      // initializing
      logdens = 0;
      
      logV = log(sum(exp(theta)));
      // note that logit(x/(x+y)) = log(x) - log(y)
      logitw1 = log(sum(exp(theta[2:3]))) - theta[1];
      logitw2 = theta[3] - theta[2];
      
      // if we do not use the likelihood we must restrict the total variance since we use the jeffreys prior
      if (use_likelihood == 0){
        logdens += normal_lpdf(logV | 0, 1);
      } else {
        logdens += jeffrey_logvar_lpdf(logV); // V, total variance
      }
      
      // adding prior for the three logit weights
      logdens += eval_spline_lpdf(logitw1 | knots, prior_coeffs, 1, n_knots); // w1, genetic vs residual
      logdens += eval_spline_lpdf(logitw2 | knots, prior_coeffs, 2, n_knots); // w2, dominance vs additive
      
      // jacobian from internal scale to regular scale (log/logit to V/w)
      logdens += logV + log(sum(exp(theta[2:3]))) - sum(theta);
      
      // jacobian from regular scale to variances (from V/w to var)
      logdens += calc_log_jacdet_regtovar_ad(theta);
      
      return(logdens);
      
    }
    
    // calculates the joint prior for theta (not including covariates) for model ADX
    real prior_adx_exp_lpdf(real[] theta, vector knots, matrix[] prior_coeffs, int n_knots, int use_likelihood){
      
      real logdens;
      real logV;
      real logitw1;
      real logitw2;
      real logitw3;
      
      // initializing
      logdens = 0;
      
      logV = log(sum(exp(theta)));
      // note that logit(x/(x+y)) = log(x) - log(y)
      logitw1 = log(sum(exp(theta[2:4]))) - theta[1];
      logitw2 = log(sum(exp(theta[3:4]))) - theta[2];
      logitw3 = theta[4] - theta[3];
      
      // if we do not use the likelihood we must restrict the total variance since we use the jeffreys prior
      if (use_likelihood == 0){
        logdens += normal_lpdf(logV | 0, 1);
      } else {
        logdens += jeffrey_logvar_lpdf(logV); // V, total variance
      }
      
      // adding prior for the three logit weights
      logdens += eval_spline_lpdf(logitw1 | knots, prior_coeffs, 1, n_knots); // w1, residual vs genetic
      logdens += eval_spline_lpdf(logitw2 | knots, prior_coeffs, 2, n_knots); // w2, effect 1 vs effect 2 and 3
      logdens += eval_spline_lpdf(logitw3 | knots, prior_coeffs, 3, n_knots); // w3, effect 2 vd effect 3
      
      // jacobian from internal scale to regular scale (log/logit to V/w)
      logdens += logV + log(sum(exp(theta[2:4]))) + log(sum(exp(theta[3:4]))) - sum(theta);
      
      // jacobian from regular scale to variances (from V/w to var)
      logdens += calc_log_jacdet_regtovar_adx(theta);
      
      return(logdens);
      
    }
    
    real prior_a_def_lpdf(real[] x){
      
      real logdens; // log density
      real alpha = 1;
      
      logdens = alpha*(x[1]+x[2]) - 2*alpha*log(exp(x[1]) + exp(x[2]));
      
      return logdens;
      
    }
    
}

data {
    
    int<lower=1> n; // number of observations (one observation per individ)
    
    real y[n]; // data
    
    int n_in_likelihood; // number of observations to keep in of likelihood
    int which_in_likelihood[n_in_likelihood]; // which observations to keep in of likelihood
    
    int<lower=1,upper=3> model_number; // which model: A = 1, AD = 2, ADX = 3
    // the number of parameters is model_number+1 (2 for A, 3 for AD, 4 for ADX)
    
    // Cholesky lower triangle matrix of the covariance matrix for the genetic effect
    matrix[n, n] L[model_number]; // A, D, X
    // matrix[n, n] L_a;
    // matrix[n, n] L_d;
    // matrix[n, n] L_e;
    
    int use_likelihood; // whether to include likelihood or not (not = sample from prior)
    
    // for the prior distribution splines
    int n_knots; // number of knots for the splines
    matrix[n_knots, 4] prior_coeffs[model_number]; // matrix with coefficients for the spline (a + bx + cx^2 + dx^3)
    vector[n_knots] knots; // the knots (logit weight scale) used for the spline (x-values)
    
  }

parameters {
  
    real intercept; // the intercept
    real theta[model_number+1]; // internal parameters, two, three or four log variances (the order is eps, A, D, X)

    // iid effect to be multiplied with the Cholesky lower triangular matrix of A, D and X
    // length depends on the model
    vector[n*model_number] u;
    
    // vector[n_z] u_a; // additive
    // vector[n_z] u_d; // dominance
    // vector[n_z] u_e; // epistasis
}

transformed parameters{
    
    //////// declaring variables
    
    // Linear predictor
    vector[n] eta;
    
    vector[n*model_number] eff;
    
    // the genetic effects, to be multiplied with the Cholesky lower triangular matrices and then centered 
    vector[n*model_number] eff_zero;
    
    // genetic effects
    eff_zero[1:n] = L[1,,] * u[1:n]; // additive effect
    eff_zero[1:n] = eff_zero[1:n] - mean(eff_zero[1:n]);
    if (model_number >= 2){
      eff_zero[(n+1):(2*n)] = L[2,,] * u[(n+1):(2*n)]; // dominance effect
      eff_zero[(n+1):(2*n)] = eff_zero[(n+1):(2*n)] - mean(eff_zero[(n+1):(2*n)]);
    }
    if (model_number == 3){
      eff_zero[(2*n+1):(3*n)] = L[3,,] * u[(2*n+1):(3*n)]; // epistasis effect
      eff_zero[(2*n+1):(3*n)] = eff_zero[(2*n+1):(3*n)] - mean(eff_zero[(2*n+1):(3*n)]);
    }
    
    // Linear predictor 
    // intercept
    eta = rep_vector(intercept, n);
    
    for (ind in 1:(num_elements(theta)-1)){
      if (exp(theta[ind+1]) > 1e-300) eta += exp(theta[ind+1]/2) * eff_zero[(1+n*(ind-1)):(n+n*(ind-1))];
    }

}

model {

    // Prior on intercept
    target += normal_lpdf(intercept | 0, sqrt(1000));
    
    // prior
    if (model_number == 1){ // model A-MW
      target += prior_a_def_lpdf(theta);
    } else if (model_number == 2){ // model AD-MW-exp
      target += prior_ad_exp_lpdf(theta | knots, prior_coeffs, n_knots, use_likelihood);
    } else { // model ADX-MW-exp
      target += prior_adx_exp_lpdf(theta | knots, prior_coeffs, n_knots, use_likelihood);
    } 
    
    target += normal_lpdf(u | 0, 1); // sampling iid effect for the genetic effects
    
    // Likelihood
    if (use_likelihood){
        // Observation equation
        // Only include the observations in the vector which_in_likelihood
        target += normal_lpdf(y[which_in_likelihood] | eta[which_in_likelihood], exp(theta[1]/2));
    }
    
}

generated quantities{
  
    vector[n] genetic_effects[model_number];
    real log_pred_dens;
  
    // genetic effects
    for (ind in 1:(num_elements(theta)-1)){
      genetic_effects[ind,] = exp(theta[ind+1]/2) * eff_zero[(1+n*(ind-1)):(n+n*(ind-1))];
    }

    // calculate log score
    log_pred_dens = normal_lpdf(y[which_in_likelihood] | eta[which_in_likelihood], exp(theta[1]/2));

}


